# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Docker shell helper class module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from topology.platforms.shell import PExpectShell, PExpectBashShell


class DockerExecMixin(object):
    """
    Docker ``exec`` connection mixin for the Topology shell API.

    This class implements a ``_get_connect_command()`` method that allows to
    interact with a shell through a ``docker exec`` interactive command, and
    extends the constructor to request for container related parameters.

    :param str container: Container unique identifier.
    :param str command: Command to be executed with the ``docker exec`` that
     will launch an interactive session.
    """

    def __init__(self, container, command, *args, **kwargs):
        self._container = container
        self._command = command
        super(DockerExecMixin, self).__init__(*args, **kwargs)

    def _get_connect_command(self):
        return 'docker exec -i -t {} {}'.format(
            self._container, self._command
        )


class DockerShell(DockerExecMixin, PExpectShell):
    """
    Generic ``docker exec`` shell for unspecified interactive session.
    """


class DockerBashShell(DockerExecMixin, PExpectBashShell):
    """
    Specialized ``docker exec`` shell that will run and setup a bash
    interactive session.
    """


__all__ = ['DockerShell', 'DockerBashShell']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Simple switch Topology Docker Node using Ubuntu.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from topology_docker.node import DockerNode
from topology_docker.shell import DockerBashShell


class SwitchNode(DockerNode):
    """
    Simple switch node for the Topology Docker platform engine.

    This base switch loads an ubuntu image (by default) and creates a
    kernel-level bridge where it adds all of its front-panel interfaces to
    emulate the behaviour of a real "dumb" switch.

    See :class:`topology_docker.node.DockerNode`.
    """

    def __init__(self, identifier, image='ubuntu:14.04', **kwargs):

        super(SwitchNode, self).__init__(identifier, image=image, **kwargs)

        # Create and register shells
        self._register_shell(
            'bash',
            DockerBashShell(
                self.container_id,
                'bash'
            )
        )
        self._register_shell(
            'bash_front_panel',
            DockerBashShell(
                self.container_id,
                'ip netns exec front_panel bash'
            )
        )

    def notify_post_build(self):
        """
        Get notified that the post build stage of the topology build was
        reached.

        See :meth:`DockerNode.notify_post_build` for more information.
        """
        super(SwitchNode, self).notify_post_build()

        # Let's create the bridge interface and then bring it up
        netns_exec = 'ip netns exec front_panel'
        self._docker_exec(
            '{} ip link add name bridge0 type bridge'.format(netns_exec)
        )
        self._docker_exec(
            '{} ip link set bridge0 up'.format(netns_exec)
        )

        # Get the ports created by the framework
        created_ports = list(self.ports.keys())

        # Add the ports to the bridge (for this they need to be UP first)
        for port in created_ports:
            self._docker_exec('{} ip link set {} up'.format(netns_exec, port))
            self._docker_exec('{} ip link set {} master bridge0'.format(
                netns_exec,
                port
            ))


__all__ = ['SwitchNode']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2018 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology_docker utilities module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from warnings import warn
from os import getuid, devnull, makedirs
from shlex import split as shsplit
from subprocess import call, check_call

log = logging.getLogger(__name__)


def ensure_dir(path):
    """
    Ensure that a path exists.

    :param str path: Directory path to create.
    """
    import errno
    EEXIST = getattr(errno, 'EEXIST', 0)  # noqa

    try:
        makedirs(path)
    except OSError as err:
        # 0 for Jython/Win32
        if err.errno not in [0, EEXIST]:
            raise


IFNAMSIZ = 15
"""
This is the maximum length a network interface can have.
Defined in linux/if.h as 16, but counting the termination character.
"""


def tmp_iface():
    """
    Return a valid temporal interface name.

    :rtype: str
    :return: A random valid network interface name.
    """
    import random
    import string
    return ''.join(
        random.choice(string.ascii_lowercase)
        for _ in range(IFNAMSIZ)
    )


def cmd_prefix():
    """
    Determine if the current user can run privileged commands and thus
    determines the command prefix to use.

    :rtype: str
    :return: The prefix to use for privileged commands.
    """

    # Return cache if set
    if hasattr(cmd_prefix, 'prefix'):
        return cmd_prefix.prefix

    if getuid() == 0:
        warn(
            'Running as root user is not recommended due to the '
            'Principle of Least Privilege. '
            'Please configure the ip command for root execution.'
        )
        cmd_prefix.prefix = ''
        return cmd_prefix.prefix

    with open(devnull, 'w') as f:
        cmd = shsplit('sudo --non-interactive ip link show')
        if call(cmd, stdout=f, stderr=f) != 0:
            raise RuntimeError(
                'Please configure the ip command for root execution.'
            )

    cmd_prefix.prefix = 'sudo --non-interactive '
    return cmd_prefix.prefix


def privileged_cmd(commands_tpl, **kwargs):
    """
    Run a privileged command.

    This function will replace the tokens in the template with the provided
    keyword arguments, then it will split the lines of the template, strip
    them and execute them using the prefix for privileged commands, checking
    that the command returns 0.

    :param str commands_tpl: Single line or multiline template for commands.
    :param dict kwargs: Replacement tokens.
    """
    prefix = cmd_prefix()

    for command in commands_tpl.format(**kwargs).splitlines():
        command = command.strip()
        if command:
            check_call(shsplit(prefix + command))


def get_iface_name(enode, netname):
    """
    Get interface name inside a container

    This function will figure out the interface name inside the container for
    a given docker network.

    :param enode: The platform ("engine") node to get the iface name from.
    :param str netname: The name of the docker network to which the interface
                        we're looking for is connected.
    """
    # FIXME: there must be a better way to do this, and we're not the only ones
    # that think so, (see: https://github.com/docker/docker/issues/17064)
    docker_netconf = enode._client.inspect_container(
        enode.container_id
    )['NetworkSettings']['Networks'][netname]

    ifaces_conf = enode._docker_exec(
        'ip -o link list'
    ).strip().split('\n')

    for iface_conf in ifaces_conf:
        if docker_netconf['MacAddress'] in iface_conf:
            iface = iface_conf.split(': ')[1].split('@')[0]
            break
    else:
        raise RuntimeError(
            'Unable to find interface with MAC address '
            '{docker_netconf[MacAddress]} in netns {netns} in container '
            '{enode._container_id} with name '
            '{enode._container_name}'.format(
                **locals()
            )
        )

    return iface


__all__ = [
    'ensure_dir',
    'tmp_iface',
    'cmd_prefix',
    'privileged_cmd',
    'get_iface_name'
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2018 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Docker engine platform module for topology.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from traceback import format_exc
from collections import OrderedDict
from docker import APIClient

from topology.platforms.utils import NodeLoader
from topology.platforms.platform import BasePlatform

from .node import DockerNode
from .utils import tmp_iface, privileged_cmd
from .networks import create_docker_network, create_platform_network


log = logging.getLogger(__name__)


class DockerPlatform(BasePlatform):
    """
    Plugin to build a topology using Docker.

    See :class:`topology.platforms.platform.BasePlatform` for more information.
    """

    def __init__(self, timestamp, nmlmanager, **kwargs):

        self.node_loader = NodeLoader(
            'docker', api_version='1.0', base_class=DockerNode
        )

        self.nmlnode_node_map = OrderedDict()
        self.nmlbiport_iface_map = OrderedDict()
        self.nmlbilink_nmlbiports_map = OrderedDict()
        self.available_node_types = self.node_loader.load_nodes()

        # Create netns folder
        privileged_cmd('mkdir -p /var/run/netns')

    def pre_build(self):
        """
        See :meth:`BasePlatform.pre_build` for more information.
        """

    def add_node(self, node):
        """
        Add a new DockerNode.

        See :meth:`BasePlatform.add_node` for more information.
        """
        # Lookup for node of given type
        node_type = node.metadata.get('type', 'host')
        if node_type not in self.available_node_types:
            raise Exception('Unknown node type {}'.format(node_type))

        # Create instance of node type and start
        enode = self.available_node_types[node_type](
            node.identifier, **node.metadata
        )

        # Register node
        self.nmlnode_node_map[node.identifier] = enode

        # Start node
        enode.start()

        # Install container netns locally
        privileged_cmd(
            'ln -s /proc/{pid}/ns/net /var/run/netns/{pid}',
            pid=enode._pid
        )

        # Manage network and their netns creation
        network_handlers = {
            'docker': create_docker_network,
            'platform': create_platform_network
        }

        for category, config in enode._get_network_config()['mapping'].items():

            managed_by = config['managed_by']
            handler = network_handlers.get(managed_by, None)

            if handler is None:
                raise RuntimeError(
                    'Unknown "managed_by" handler {}'.format(managed_by)
                )

            handler(enode, category, config)

        return enode

    def add_biport(self, node, biport):
        """
        Add a port to the docker node.

        See :meth:`BasePlatform.add_biport` for more information.
        """
        enode = self.nmlnode_node_map[node.identifier]
        eport = enode.notify_add_biport(node, biport)

        # Get network configuration of given port
        network_config = enode._get_network_config()

        category = biport.metadata.get(
            'category',
            network_config['default_category']
        )
        category_config = network_config['mapping'][category]

        # If this port is from a docker-managed network, then it will be
        # created when this node was connected to the network
        created = True if category_config['managed_by'] == 'docker' else False

        # Sanity check to make sure that this biport identifier is not
        # a duplicate
        if biport.identifier in self.nmlbiport_iface_map:
            raise ValueError(
                'Biport identifier already used: {}'.format(biport.identifier)
            )

        # Register this port for later creation
        self.nmlbiport_iface_map[biport.identifier] = {
            'created': created,
            'iface_base': eport,
            'prefix': category_config['prefix'],
            'iface': '{}{}'.format(category_config['prefix'], eport),
            'container_ns': enode._pid,
            'netns': category_config['netns'],
            'owner': node.identifier,
            'label': biport.metadata.get('label', biport.identifier)
        }

        return eport

    def add_bilink(self, nodeport_a, nodeport_b, bilink):
        """
        Add a link between two nodes.

        See :meth:`BasePlatform.add_bilink` for more information.
        """
        node_a, port_a = nodeport_a
        node_b, port_b = nodeport_b

        # Get enodes
        enode_a = self.nmlnode_node_map[node_a.identifier]
        enode_b = self.nmlnode_node_map[node_b.identifier]

        # Determine temporal interfaces names
        tmp_iface_a = tmp_iface()
        tmp_iface_b = tmp_iface()

        # Get port spec dictionary
        port_spec_a = self.nmlbiport_iface_map[port_a.identifier]
        port_spec_b = self.nmlbiport_iface_map[port_b.identifier]

        # If any of the ports is already created then don't do anything here
        # FIXME: We should probably let the user know if one of the ports is
        # created and the other is not, as this case is undefined and
        # unsupported
        if port_spec_a['created'] or port_spec_b['created']:
            return

        # Determine final interface names
        iface_a = port_spec_a['iface']
        iface_b = port_spec_b['iface']

        # Create links between nodes:
        #   docs.docker.com/v1.5/articles/networking/#building-a-point-to-point-connection # noqa
        commands = """\
        ip link add {tmp_iface_a} type veth peer name {tmp_iface_b}
        ip link set {tmp_iface_a} netns {enode_a._pid} name {iface_a}
        ip link set {tmp_iface_b} netns {enode_b._pid} name {iface_b}\
        """
        privileged_cmd(commands, **locals())

        # Apply some attributes to nodes and ports
        for enode, port, port_spec, iface in (
            (enode_a, port_a, port_spec_a, iface_a),
            (enode_b, port_b, port_spec_b, iface_b)
        ):

            # Move interfaces to correct network namespace
            netns = port_spec['netns']
            if netns:
                enode._docker_exec(
                    'ip link set dev {iface} netns {netns}'.format(
                        **locals()
                    )
                )
                cmd_prefix = 'ip netns exec {netns} '.format(**locals())
            else:
                cmd_prefix = ''

            # Set ipv4 and ipv6 addresses
            for version in [4, 6]:
                attribute = 'ipv{}'.format(version)
                if attribute not in port.metadata:
                    continue

                addr = port.metadata[attribute]
                cmd = (
                    '{cmd_prefix}ip -{version} addr add {addr} dev {iface}'
                ).format(**locals())
                enode._docker_exec(cmd)

            # Bring-up or down
            if bilink.metadata.get('up', None) is None and \
                    port.metadata.get('up', None) is None:
                continue

            up = bilink.metadata.get('up', True) and \
                port.metadata.get('up', True)

            state = 'up' if up else 'down'
            cmd = '{cmd_prefix}ip link set dev {iface} {state}'.format(
                **locals()
            )
            enode._docker_exec(cmd)

        # Notify enodes of created interfaces
        enode_a.notify_add_bilink(nodeport_a, bilink)
        enode_b.notify_add_bilink(nodeport_b, bilink)

        # Mark interfaces as created
        self.nmlbiport_iface_map[port_a.identifier]['created'] = True
        self.nmlbiport_iface_map[port_b.identifier]['created'] = True

        # Register these links
        self.nmlbilink_nmlbiports_map[bilink.identifier] = (
            port_a.identifier, port_b.identifier
        )

    def post_build(self):
        """
        Ports are created for each node automatically while adding links.
        Creates the rest of the ports (no-linked ports)

        See :meth:`BasePlatform.post_build` for more information.
        """
        # Create remaining interfaces
        cmd_tpl = (
            'ip netns exec {container_ns} '
            'ip tuntap add dev {iface} mode tap'
        )
        for port_spec in self.nmlbiport_iface_map.values():

            # Ignore already created interfaces
            if port_spec['created']:
                continue

            # Create port as dummy tuntap device
            privileged_cmd(cmd_tpl, **port_spec)

            # Move port to network namespace if required
            if port_spec['netns']:
                enode = self.nmlnode_node_map[port_spec['owner']]
                enode._docker_exec(
                    'ip link set dev {iface} netns {netns}'.format(
                        **port_spec
                    )
                )

            # Mark as created
            port_spec['created'] = True

        # Notify nodes of the post_build event
        for enode in self.nmlnode_node_map.values():
            enode.notify_post_build()

    def destroy(self):
        """
        See :meth:`BasePlatform.destroy` for more information.
        """
        # NOTE: Implementation is split on purpose

        # Request termination of all containers
        for enode in self.nmlnode_node_map.values():
            try:
                enode.stop()
            except Exception:
                log.error(format_exc())

        # Remove the linked netns
        for enode in self.nmlnode_node_map.values():
            try:
                privileged_cmd('rm /var/run/netns/{pid}', pid=enode._pid)
            except Exception:
                log.error(format_exc())

        # Save the names of all docker-managed networks
        networks_to_remove = set()
        for enode in self.nmlnode_node_map.values():
            try:
                network_config = enode._get_network_config()['mapping']

                for category, config in network_config.items():
                    if config['managed_by'] == 'docker':
                        netname = config.get('connect_to', None)
                        if netname is None:

                            netname = '{}_{}'.format(
                                enode._container_name,
                                category
                            )

                        networks_to_remove.add(netname)

            except Exception:
                log.error(format_exc())

        # Remove all docker-managed networks
        dockerclient = APIClient(version='auto')
        for netname in networks_to_remove:
            dockerclient.remove_network(net_id=netname)

    def rollback(self, stage, enodes, exception):
        """
        See :meth:`BasePlatform.rollback` for more information.
        """
        self.destroy()

    def _common_link(self, link_id, action):
        """
        Common action to relink / unlink.

        :param str link_id: Identifier of the link to modify.
        :param bool action: True if up, False if down.
        """
        if link_id not in self.nmlbilink_nmlbiports_map:
            raise Exception('Unknown link "{}"'.format(link_id))

        # iterate endpoints
        for port_id in self.nmlbilink_nmlbiports_map[link_id]:

            # Get specification for this endpoint
            port_spec = self.nmlbiport_iface_map[port_id]

            # Get node for the owner of this port
            enode = self.nmlnode_node_map[port_spec['owner']]
            enode.set_port_state(port_spec['label'], action)

    def relink(self, link_id):
        """
        See :meth:`BasePlatform.relink` for more information.
        """
        self._common_link(link_id, True)

    def unlink(self, link_id):
        """
        See :meth:`BasePlatform.unlink` for more information.
        """
        self._common_link(link_id, False)


__all__ = ['DockerPlatform']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Simple host Topology Docker Node using Ubuntu.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from topology_docker.node import DockerNode
from topology_docker.shell import DockerBashShell


class HostNode(DockerNode):
    """
    Simple host node for the Topology Docker platform engine.

    This base host loads an ubuntu image (by default) and has bash as the
    default shell.

    See :class:`topology_docker.node.DockerNode`.
    """

    def __init__(self, identifier, image='ubuntu:14.04', **kwargs):

        super(HostNode, self).__init__(identifier, image=image, **kwargs)

        # Create and register shells
        self._register_shell(
            'bash',
            DockerBashShell(
                self.container_id,
                'ip netns exec front_panel bash'
            )
        )
        self._register_shell(
            'bash_oobm',
            DockerBashShell(
                self.container_id,
                'bash'
            )
        )


__all__ = ['HostNode']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2018 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology_docker base node module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from os import getpid
from json import loads
from os.path import join
from logging import getLogger
from datetime import datetime
from shlex import split as shsplit
from subprocess import check_output
from abc import ABCMeta, abstractmethod

from docker import APIClient
from six import add_metaclass

from topology.platforms.node import CommonNode
from topology_docker.utils import ensure_dir, get_iface_name


log = getLogger(__name__)


@add_metaclass(ABCMeta)
class DockerNode(CommonNode):
    """
    An instance of this class will create a detached Docker container.

    This node binds the ``shared_dir_mount`` directory of the container to a
    local path in the host system defined in ``self.shared_dir``.

    :param str identifier: Node unique identifier in the topology being built.
    :param str image: The image to run on this node, in the
     form ``repository:tag``.
    :param str registry: Docker registry to pull image from.
    :param str command: The command to run when the container is brought up.
    :param str binds: Directories to bind for this container separated by a
     ``;`` in the form:

     ::

        '/tmp:/tmp;/dev/log:/dev/log;/sys/fs/cgroup:/sys/fs/cgroup'

    :param str network_mode: Network mode for this container.
    :param str hostname: Container hostname.
    :param environment: Environment variables to pass to the
     container. They can be set as a list of strings in the following format:

     ::

        ['environment_variable=value']

     or as a dictionary in the following format:

     ::

        {'environment_variable': 'value'}

    :type environment: list or dict
    :param bool privileged: Run container in privileged mode or not.
    :param bool tty: Whether to allocate a TTY or not to the process.
    :param str shared_dir_base: Base path in the host where the shared
     directory will be created. The shared directory will always have the name
     of the container inside this directory.
    :param str shared_dir_mount: Mount point of the shared directory in the
     container.
    :param dict create_host_config_kwargs: Extra kwargs arguments to pass to
     docker-py's ``create_host_config()`` low-level API call.
    :param dict create_container_kwargs: Extra kwargs arguments to pass to
     docker-py's ``create_container()`` low-level API call.

    Read only public attributes:

    :var str image: Name of the Docker image being used by this node.
     Same as the ``image`` keyword argument.
    :var str container_id: Unique container identifier assigned by the Docker
     daemon in the form of a hash.
    :var str container_name: Unique container name assigned by the framework in
     the form ``{identifier}_{pid}_{timestamp}``.
    :var str shared_dir: Share directory in the host for this container. Always
     ``/tmp/topology/{container_name}``.
    :var str shared_dir_mount: Directory inside the container where the
     ``shared_dir`` is mounted. Same as the ``shared_dir_mount`` keyword

    .. automethod:: _get_network_config
    """

    @abstractmethod
    def __init__(
            self, identifier,
            image='ubuntu:latest', registry=None, command='bash',
            binds=None, network_mode='none', hostname=None,
            environment=None, privileged=True, tty=True,
            shared_dir_base='/tmp/topology/docker/',
            shared_dir_mount='/var/topology',
            create_host_config_kwargs=None,
            create_container_kwargs=None,
            **kwargs):

        super(DockerNode, self).__init__(identifier, **kwargs)

        self._pid = None
        self._image = image
        self._registry = registry
        self._command = command
        self._hostname = hostname
        self._environment = environment
        self._client = APIClient(version='auto')

        self._container_name = '{identifier}_{pid}_{timestamp}'.format(
            identifier=identifier.replace('>', '.'), pid=getpid(),
            timestamp=datetime.now().isoformat().replace(':', '-')
        )
        self._shared_dir_base = shared_dir_base
        self._shared_dir_mount = shared_dir_mount
        self._shared_dir = join(
            shared_dir_base,
            self._container_name
        )

        self._create_host_config_kwargs = create_host_config_kwargs or {}
        self._create_container_kwargs = create_container_kwargs or {}

        # Autopull docker image if necessary
        self._autopull()

        # Create shared directory
        ensure_dir(self._shared_dir)

        # Add binded directories
        container_binds = [
            '{}:{}'.format(
                self._shared_dir,
                self._shared_dir_mount
            )
        ]
        if binds is not None:
            container_binds.extend(binds.split(';'))

        # Create host config
        create_host_config_call = {
            'privileged': privileged,
            'network_mode': network_mode,
            'binds': container_binds,
            'init': True
        }
        create_host_config_call.update(self._create_host_config_kwargs)

        self._host_config = self._client.create_host_config(
            **create_host_config_call
        )

        # Create container
        create_container_call = {
            'image': self._image,
            'command': self._command,
            'name': self._container_name,
            'detach': True,
            'tty': tty,
            'hostname': self._hostname,
            'host_config': self._host_config,
            'environment': self._environment,
        }
        create_container_call.update(self._create_container_kwargs)

        self._container_id = self._client.create_container(
            **create_container_call
        )['Id']

    @property
    def image(self):
        return self._image

    @property
    def container_id(self):
        return self._container_id

    @property
    def container_name(self):
        return self._container_name

    @property
    def shared_dir(self):
        return self._shared_dir

    @property
    def shared_dir_mount(self):
        return self._shared_dir_mount

    def _get_network_config(self):
        """
        Defines the network configuration for nodes of this type.

        This method should be overriden when implementing a new node type to
        return a dictionary with its network configuration by setting the
        following components:

        'mapping'
            This is a dictionary of dictionaries, each parent-level key defines
            one network category, and each category *must* have these three
            keys: **netns**, **managed_by**, and **prefix**, and *can*
            (optionally) have a **connect_to** key).

            'netns'
                Specifies the network namespace (inside the docker container)
                where all the ports belonging to this category will be moved
                after their creation. If set to None, then the ports will
                remain in the container's default network namespace.

            'managed_by'
                Specifies who will manage different aspects of this network
                category depending on its value (which can be either **docker**
                or **platform**).

                'docker'
                    This network category will represent a network created by
                    docker (identical to using the docker network create
                    command) and will be visible to docker (right now all
                    docker-managed networks are created using docker's 'bridge'
                    built-in network plugin, this will likely change in the
                    near future).

                'platform'
                    This network category will represent ports created by the
                    Docker Platform Engine and is invisible to docker.

            'prefix'
                Defines a prefix that will be used when a port/interface is
                moved into a namespace, its value can be set to '' (empty
                string) if no prefix is needed. In cases where the parent
                network category doesn't have a netns (i.e. 'netns' is set to
                None) this value will be ignored.

            'connect_to'
                Specifies a Docker network this category will be connected to,
                if this network doesn't exists it will be created. If set to
                None, this category will be connected to a uniquely named
                Docker network that will be created by the platform.

        'default_category'
            Every port that didn't explicitly set its category (using the
            "category" attribute in the SZN definition) will be set to this
            category.

        This is an example of a network configuration dictionary as expected to
        be returned by this funcition::

            {
                'default_category': 'front_panel',
                'mapping': {
                    'oobm': {
                        'netns': 'oobmns',
                        'managed_by': 'docker',
                        'connect_to': 'oobm'
                        'prefix': ''
                    },
                    'back_panel': {
                        'netns': None,
                        'managed_by': 'docker',
                        'prefix': ''
                    },
                    'front_panel': {
                        'netns': 'front',
                        'managed_by': 'platform',
                        'prefix': 'f_'
                    }
                }
            }

        :returns: The dictionary defining the network configuration.
        :rtype: dict
        """
        return {
            'default_category': 'front_panel',
            'mapping': {
                'oobm': {
                    'netns': None,
                    'managed_by': 'docker',
                    'prefix': ''
                },
                'front_panel': {
                    'netns': 'front_panel',
                    'managed_by': 'platform',
                    'prefix': ''
                }
            }
        }

    def _autopull(self):
        """
        Autopulls the docker image of the node, if necessary.
        """
        # Search for image in available images
        for tags in [img['RepoTags'] for img in self._client.images()]:
            # Docker py can return repo tags as None
            if tags and self._image in tags:
                return

        # Determine image parts
        registry = self._registry
        image = self._image
        tag = 'latest'

        if ':' in image:
            image, tag = image.split(':')

        # Pull image
        pull_uri = image
        if registry:
            pull_uri = '{}/{}'.format(registry, image)
        pull_name = '{}:{}'.format(pull_uri, tag)

        log.info('Trying to pull image {} ...'.format(pull_name))

        last = ''
        for line in self._client.pull(pull_uri, tag=tag, stream=True):
            last = line
        status = loads(last.decode('utf8'))

        log.debug('Pulling result :: {}'.format(status))

        if 'error' in status:
            raise Exception(status['error'])

        # Retag if required
        if pull_name != self._image:
            if not self._client.tag(pull_name, image, tag):
                raise Exception(
                    'Error when tagging image {} with tag {}:{}'.format(
                        pull_name, image, tag
                    )
                )

            log.info(
                'Tagged image {} with tag {}:{}'.format(
                    pull_name, image, tag
                )
            )

    def _docker_exec(self, command):
        """
        Execute a command inside the docker.

        :param str command: The command to execute.
        """
        log.debug(
            '[{}]._docker_exec(\'{}\') ::'.format(self._container_id, command)
        )

        response = check_output(shsplit(
            'docker exec {container_id} {command}'.format(
                container_id=self._container_id, command=command.strip()
            )
        )).decode('utf8')

        log.debug(response)
        return response

    def _get_services_address(self):
        """
        Get the service address of the node using Docker's inspect mechanism
        to grab OOBM interface address.

        :return: The address (IP or FQDN) of the services interface (oobm).
        :rtype: str
        """
        network_name = self._container_name + '_oobm'
        address = self._client.inspect_container(
            self.container_id
        )['NetworkSettings']['Networks'][network_name]['IPAddress']
        return address

    def notify_add_biport(self, node, biport):
        """
        Get notified that a new biport was added to this engine node.

        :param node: The specification node that spawn this engine node.
        :type node: pynml.nml.Node
        :param biport: The specification bidirectional port added.
        :type biport: pynml.nml.BidirectionalPort
        :rtype: str
        :return: The assigned interface name of the port.
        """

        network_config = self._get_network_config()

        category = biport.metadata.get(
            'category',
            network_config['default_category']
        )
        category_config = network_config['mapping'][category]

        if category_config['managed_by'] == 'docker':
            netname = category_config.get(
                'connect_to',
                '{}_{}'.format(self._container_name, category)
            )
            return get_iface_name(self, netname)
        else:
            return biport.metadata.get('label', biport.identifier)

    def notify_add_bilink(self, nodeport, bilink):
        """
        Get notified that a new bilink was added to a port of this engine node.

        :param nodeport: A tuple with the specification node and port being
         linked.
        :type nodeport: (pynml.nml.Node, pynml.nml.BidirectionalPort)
        :param bilink: The specification bidirectional link added.
        :type bilink: pynml.nml.BidirectionalLink
        """

    def notify_post_build(self):
        """
        Get notified that the post build stage of the topology build was
        reached.
        """
        # Log container data
        image_data = self._client.inspect_image(
            image=self._image
        )
        log.info(
            'Started container {}:\n'
            '    Image name: {}\n'
            '    Image id: {}\n'
            '    Image creation date: {}'
            '    Image tags: {}'.format(
                self._container_name,
                self._image,
                image_data.get('Id', '????'),
                image_data.get('Created', '????'),
                ', '.join(image_data.get('RepoTags', []))
            )
        )
        container_data = self._client.inspect_container(
            container=self._container_id
        )
        log.debug(container_data)

    def start(self):
        """
        Start the docker node and configures a netns for it.
        """
        self._client.start(self._container_id)
        self._pid = self._client.inspect_container(
            self._container_id)['State']['Pid']

    def stop(self):
        """
        Request container to stop.
        """
        self._client.stop(self._container_id)
        self._client.wait(self._container_id)
        self._client.remove_container(self._container_id)

    def disable(self):
        """
        Disable the node.

        In Docker implementation this pauses the container.
        """
        for portlbl in self.ports:
            self.set_port_state(portlbl, False)
        self._client.pause(self._container_id)

    def enable(self):
        """
        Enable the node.

        In Docker implementation this unpauses the container.
        """
        self._client.unpause(self._container_id)
        for portlbl in self.ports:
            self.set_port_state(portlbl, True)

    def set_port_state(self, portlbl, state):
        """
        Set the given port label to the given state.

        :param str portlbl: The label of the port.
        :param bool state: True for up, False for down.
        """
        iface = self.ports[portlbl]
        state = 'up' if state else 'down'

        command = (
            'ip netns exec front_panel '
            'ip link set dev {iface} {state}'.format(**locals())
        )
        self._docker_exec(command)


__all__ = ['DockerNode']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2020 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology_docker module entry point.
"""

__author__ = 'Hewlett Packard Enterprise Development LP'
__email__ = 'hpe-networking@lists.hp.com'
__version__ = '1.8.2'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Docker networks management module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from topology_docker.utils import get_iface_name


log = logging.getLogger(__name__)


def create_docker_network(enode, category, config):
    """
    Create a Docker managed network with given configuration (netns, prefix) to
    be used with Topology framework.

    :param enode: The platform (a.k.a "engine") node to configure.
    :param str category: Name of the panel category.
    :param dict config: Configuration for the network. A dictionary like:

     ::

        {
            'netns': 'mynetns',
            'managed_by': 'docker',
            'connect_to': 'somedockernetwork',
            'prefix': ''
        }

     This dictionary is taken from the ``node._get_network_config()`` result.
    """
    # Let's find out which networks exist already
    dockernets = enode._client.networks()

    # Let's figure out what docker network we should connect to
    netname = config.get(
        'connect_to',
        '{}_{}'.format(enode._container_name, category)
    )

    # Create docker network if it doesn't exist already
    if not any(d['Name'] == netname for d in dockernets):
        enode._client.create_network(
            name=netname,
            driver='bridge'
        )

    # Disconnect from 'none' to be able to connect to other
    # networks (https://github.com/docker/docker/issues/21132)
    networks = enode._client.inspect_container(
        enode.container_id
    )['NetworkSettings']['Networks']
    if 'none' in networks:
        enode._client.disconnect_container_from_network(
            container=enode._container_id,
            net_id='none'
        )

    # Connect container to the docker network
    enode._client.connect_container_to_network(
        container=enode._container_id,
        net_id=netname
    )

    # Check if this category has a defined netns
    netns = config.get('netns', None)
    if netns is None:
        return

    # Create this network's namespace inside the container
    # https://imgflip.com/i/16621d
    enode._docker_exec('ip netns add {}'.format(netns))

    netns_exec = 'ip netns exec {}'.format(netns)

    # lo should always be up
    enode._docker_exec('{} ip link set dev lo up'.format(netns_exec))

    # Find out the name Docker gave to this interface name
    iface = get_iface_name(enode, netname)

    # Prefix interface
    prefixed_iface = '{}{}'.format(config['prefix'], iface)

    # Move this network's interface to its netns
    enode._docker_exec(
        'ip link set dev {iface} netns {netns} name {prefixed_iface}'.format(
            **locals()
        )
    )

    # Reset the interface to original config
    # This is required because after moving the iface from netns it lost its
    # ip and other config.
    enode._docker_exec(
        '{netns_exec} '
        'ip address add {docker_netconf[IPAddress]}/'
        '{docker_netconf[IPPrefixLen]} '
        'dev {prefixed_iface}'.format(
            **locals()
        )
    )
    enode._docker_exec(
        '{netns_exec} ip link set dev {prefixed_iface} up'.format(
            **locals()
        )
    )


def create_platform_network(enode, category, config):
    """
    Create a Topology managed network with given configuration (netns, prefix).

    :param enode: The platform (a.k.a "engine") node to configure.
    :param str category: Name of the panel category.
    :param dict config: Configuration for the network. A dictionary like:

     ::

        {
            'netns': 'mynetns',
            'managed_by': 'platform',
            'prefix': ''
        }

     This dictionary is taken from the ``node._get_network_config()`` result.
    """
    # Check if this category has a defined netns
    netns = config.get('netns', None)
    if netns is None:
        return

    # Create the given network namespace
    enode._docker_exec('ip netns add {}'.format(netns))

    # lo should always be up
    enode._docker_exec('ip netns exec {} ip link set dev lo up'.format(netns))


__all__ = ['create_docker_network', 'create_platform_network']
